@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2002-2003 Ardis Technolgies <roman@ardistech.com>
+ * Copyright (C) 2008 Arne Redlich <agr@powerkom-dd.de>
  *
  * Released under the terms of the GNU GPL v2.0.
  */
@@ -7,6 +8,7 @@
 #ifndef __ISCSI_H__
 #define __ISCSI_H__
 
+#include <linux/blkdev.h>
 #include <linux/completion.h>
 #include <linux/pagemap.h>
 #include <linux/seq_file.h>
@@ -17,6 +19,7 @@
 
 #include "iscsi_hdr.h"
 #include "iet_u.h"
+#include "compat.h"
 
 #define IET_SENSE_BUF_SIZE      18
 
@@ -46,6 +49,8 @@ struct iscsi_trgt_param {
 	int wthreads;
 	int target_type;
 	int queued_cmnds;
+	int nop_interval;
+	int nop_timeout;
 };
 
 struct tio {
@@ -89,6 +94,8 @@ struct worker_thread_info {
 	struct list_head work_queue;
 
 	wait_queue_head_t wthread_sleep;
+
+	struct io_context *wthread_ioc;
 };
 
 struct iscsi_cmnd;
@@ -124,7 +131,6 @@ struct iscsi_target {
 	struct worker_thread_info * wthread_info;
 
 	struct semaphore target_sem;
-	struct completion *done;
 };
 
 struct iscsi_queue {
@@ -146,7 +152,7 @@ struct iet_volume {
 	struct iscsi_queue queue;
 
 	u8 scsi_id[SCSI_ID_LEN];
-	u8 scsi_sn[SCSI_SN_LEN];
+	u8 scsi_sn[SCSI_SN_LEN + 1];
 
 	u32 blk_shift;
 	u64 blk_cnt;
@@ -212,6 +218,7 @@ enum connection_state_bit {
 	CONN_ACTIVE,
 	CONN_CLOSING,
 	CONN_WSPACE_WAIT,
+	CONN_NEED_NOP_IN,
 };
 
 #define ISCSI_CONN_IOV_MAX	(((256 << 10) >> PAGE_SHIFT) + 1)
@@ -238,6 +245,7 @@ struct iscsi_conn {
 	atomic_t nr_busy_cmnds;
 	struct list_head pdu_list;		/* in/outcoming pdus */
 	struct list_head write_list;		/* list of data pdus to be sent */
+	struct timer_list nop_timer;
 
 	struct iscsi_cmnd *read_cmnd;
 	struct msghdr read_msg;
@@ -284,6 +292,8 @@ struct iscsi_cmnd {
 
 	u8 status;
 
+	struct timer_list timer;
+
 	u32 r2t_sn;
 	u32 r2t_length;
 	u32 is_unsolicited_data;
@@ -323,12 +333,12 @@ extern void send_data_rsp(struct iscsi_cmnd *, void (*)(struct iscsi_cmnd *));
 extern void send_scsi_rsp(struct iscsi_cmnd *, void (*)(struct iscsi_cmnd *));
 extern void iscsi_cmnd_set_sense(struct iscsi_cmnd *, u8 sense_key, u8 asc,
 				 u8 ascq);
+extern void send_nop_in(struct iscsi_conn *);
 
 /* conn.c */
 extern struct iscsi_conn *conn_lookup(struct iscsi_session *, u16);
 extern int conn_add(struct iscsi_session *, struct conn_info *);
 extern int conn_del(struct iscsi_session *, struct conn_info *);
-extern void conn_del_all(struct iscsi_session *);
 extern int conn_free(struct iscsi_conn *);
 extern void conn_close(struct iscsi_conn *);
 extern void conn_info_show(struct seq_file *, struct iscsi_session *);
@@ -369,7 +379,6 @@ extern struct file_operations session_seq_fops;
 extern struct iscsi_session *session_lookup(struct iscsi_target *, u64);
 extern int session_add(struct iscsi_target *, struct session_info *);
 extern int session_del(struct iscsi_target *, u64);
-extern void session_del_all(struct iscsi_target *);
 
 /* volume.c */
 extern struct file_operations volume_seq_fops;
@@ -456,8 +465,11 @@ static inline void iscsi_cmnd_set_length(struct iscsi_pdu *pdu)
 #define cmnd_itt(cmnd) cpu_to_be32((cmnd)->pdu.bhs.itt)
 #define cmnd_opcode(cmnd) ((cmnd)->pdu.bhs.opcode & ISCSI_OPCODE_MASK)
 #define cmnd_scsicode(cmnd) cmnd_hdr(cmnd)->scb[0]
+#define cmnd_immediate(cmnd) ((cmnd)->pdu.bhs.opcode & ISCSI_OP_IMMEDIATE)
 
-#define	SECTOR_SIZE_BITS	9
+/* default and maximum scsi level block sizes */
+#define IET_DEF_BLOCK_SIZE	512
+#define IET_MAX_BLOCK_SIZE	4096
 
 enum cmnd_flags {
 	CMND_hashed,
@@ -469,6 +481,7 @@ enum cmnd_flags {
 	CMND_pending,
 	CMND_tmfabort,
 	CMND_rxstart,
+	CMND_timer_active,
 };
 
 #define set_cmnd_hashed(cmnd)	set_bit(CMND_hashed, &(cmnd)->flags)
@@ -499,6 +512,11 @@ enum cmnd_flags {
 #define set_cmnd_rxstart(cmnd)	set_bit(CMND_rxstart, &(cmnd)->flags)
 #define cmnd_rxstart(cmnd)	test_bit(CMND_rxstart, &(cmnd)->flags)
 
+#define set_cmnd_timer_active(cmnd)  set_bit(CMND_timer_active, &(cmnd)->flags)
+#define clear_cmnd_timer_active(cmnd) \
+	                        clear_bit(CMND_timer_active, &(cmnd)->flags)
+#define cmnd_timer_active(cmnd) test_bit(CMND_timer_active, &(cmnd)->flags)
+
 #define VENDOR_ID	"IET"
 #define PRODUCT_ID	"VIRTUAL-DISK"
 #define PRODUCT_REV	"0"
