@@ -82,21 +82,21 @@ retry:
 		exit(1);
 	}
 
-	log_debug(1, "conn %u session %#" PRIx64 " target %u, state %u",
+	log_debug(1, "conn %u session %llu target %u, state %u",
 		  event.cid, event.sid, event.tid, event.state);
 
 	switch (event.state) {
 	case E_CONN_CLOSE:
 		if (!(session = session_find_id(event.tid, event.sid))) {
-			log_warning("session %#" PRIx64 " not found?", event.sid);
+			log_warning("session %llu not found?", event.sid);
 			goto retry;
 		}
 
-		if (!--session->conn_cnt)
+		if (--session->conn_cnt <= 0)
 			session_remove(session);
 		break;
 	default:
-		log_warning("%s(%d) %u\n", __FUNCTION__, __LINE__, event.state);
+		log_error("%s(%d) %u\n", __FUNCTION__, __LINE__, event.state);
 		exit(-1);
 		break;
 	}
