@@ -1,7 +1,28 @@
 /*
- * Copyright (C) 2002-2003 Ardis Technolgies <roman@ardistech.com>
+ * iscsid.c - ietd iSCSI protocol processing
+ *
+ * Copyright (C) 2002-2003 Ardis Technolgies <roman at ardistech dot com>
+ * Copyright (C) 2004-2010 VMware, Inc. All Rights Reserved.
+ * Copyright (C) 2007-2010 Ross Walker <rswwalker at gmail dot com>
+ *
+ * This file is part of iSCSI Enterprise Target software.
  *
  * Released under the terms of the GNU GPL v2.0.
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
+ * 02110-1301 USA
  */
 
 #include <ctype.h>
@@ -189,9 +210,30 @@ static int account_empty(u32 tid, int dir)
 	return cops->account_query(tid, dir, pass, pass) < 0 ? 1 : 0;
 }
 
+static void login_rsp_err(struct connection *conn, int status_class,
+							int status_detail)
+{
+	struct iscsi_login_rsp_hdr *rsp =
+				(struct iscsi_login_rsp_hdr *)&conn->rsp.bhs;
+
+	rsp->flags = 0;
+	rsp->status_class = status_class;
+	rsp->status_detail = status_detail;
+	conn->state = STATE_EXIT;
+}
+
+static inline void login_rsp_ini_err(struct connection *conn, int status_detail)
+{
+	login_rsp_err(conn, ISCSI_STATUS_INITIATOR_ERR, status_detail);
+}
+
+static inline void login_rsp_tgt_err(struct connection *conn, int status_detail)
+{
+	login_rsp_err(conn, ISCSI_STATUS_TARGET_ERR, status_detail);
+}
+
 static void text_scan_security(struct connection *conn)
 {
-	struct iscsi_login_rsp_hdr *rsp = (struct iscsi_login_rsp_hdr *)&conn->rsp.bhs;
 	char *key, *value, *data, *nextValue;
 	int datasize;
 
@@ -227,18 +269,14 @@ static void text_scan_security(struct connection *conn)
 		} else
 			text_key_add(conn, key, "NotUnderstood");
 	}
-	if (conn->auth_method == AUTH_UNKNOWN) {
-		rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-		rsp->status_detail = ISCSI_STATUS_AUTH_FAILED;
-		conn->state = STATE_EXIT;
-	}
+	if (conn->auth_method == AUTH_UNKNOWN)
+		login_rsp_ini_err(conn, ISCSI_STATUS_AUTH_FAILED);
 }
 
 static void login_security_done(struct connection *conn)
 {
-	int err;
-	struct iscsi_login_req_hdr *req = (struct iscsi_login_req_hdr *)&conn->req.bhs;
-	struct iscsi_login_rsp_hdr *rsp = (struct iscsi_login_rsp_hdr *)&conn->rsp.bhs;
+	struct iscsi_login_req_hdr *req =
+				(struct iscsi_login_req_hdr *)&conn->req.bhs;
 	struct session *session;
 
 	if (!conn->tid)
@@ -247,25 +285,25 @@ static void login_security_done(struct connection *conn)
 	if ((session = session_find_name(conn->tid, conn->initiator, req->sid))) {
 		if (!req->sid.id.tsih) {
 			/* do session reinstatement */
-			session_conns_close(conn->tid, session->sid.id64);
-			session = NULL;
+			log_debug(1, "session %#" PRIx64 " reinstated",
+								req->sid.id64);
+			if (session_remove(session))
+				login_rsp_tgt_err(conn, ISCSI_STATUS_TARGET_ERROR);
+			return;
 		} else if (req->sid.id.tsih != session->sid.id.tsih) {
 			/* fail the login */
-			rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-			rsp->status_detail = ISCSI_STATUS_SESSION_NOT_FOUND;
-			conn->state = STATE_EXIT;
+			login_rsp_ini_err(conn, ISCSI_STATUS_SESSION_NOT_FOUND);
 			return;
-		} else if ((err = conn_test(conn)) == -ENOENT) {
-			/* do connection reinstatement */
 		}
-		/* add a new connection to the session */
+		/* add connection to existing session */
+		/* reinstatement handled in kernel */
+		log_debug(1, "connection %u added to session %#" PRIx64,
+						conn->cid, req->sid.id64);
 		conn->session = session;
 	} else {
 		if (req->sid.id.tsih) {
 			/* fail the login */
-			rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-			rsp->status_detail = ISCSI_STATUS_SESSION_NOT_FOUND;
-			conn->state = STATE_EXIT;
+			login_rsp_ini_err(conn, ISCSI_STATUS_SESSION_NOT_FOUND);
 			return;
 		}
 		/* instantiate a new session */
@@ -276,7 +314,6 @@ static void text_scan_login(struct connection *conn)
 {
 	char *key, *value, *data;
 	int datasize, idx;
-	struct iscsi_login_rsp_hdr *rsp = (struct iscsi_login_rsp_hdr *)&conn->rsp.bhs;
 
 	data = conn->req.data;
 	datasize = conn->req.datasize;
@@ -303,12 +340,8 @@ static void text_scan_login(struct connection *conn)
 				    == KEY_STATE_START) {
 					text_key_add_reject(conn, key);
 					continue;
-				} else {
-					rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-					rsp->status_detail = ISCSI_STATUS_INIT_ERR;
-					conn->state = STATE_EXIT;
-					goto out;
-				}
+				} else
+					goto init_err;
 			}
 
 			err = param_check_val(session_keys, idx, &val);
@@ -324,12 +357,9 @@ static void text_scan_login(struct connection *conn)
 				break;
 			case KEY_STATE_REQUEST:
 				if (val != conn->session_param[idx].val) {
-					rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-					rsp->status_detail = ISCSI_STATUS_INIT_ERR;
-					conn->state = STATE_EXIT;
 					log_warning("%s %u %u\n", key,
 					val, conn->session_param[idx].val);
-					goto out;
+					goto init_err;
 				}
 				break;
 			case KEY_STATE_DONE:
@@ -340,7 +370,9 @@ static void text_scan_login(struct connection *conn)
 			text_key_add(conn, key, "NotUnderstood");
 	}
 
-out:
+	return;
+init_err:
+	login_rsp_ini_err(conn, ISCSI_STATUS_INIT_ERR);
 	return;
 }
 
@@ -388,8 +420,9 @@ static int text_check_param(struct connection *conn)
 
 static void login_start(struct connection *conn)
 {
-	struct iscsi_login_req_hdr *req = (struct iscsi_login_req_hdr *)&conn->req.bhs;
-	struct iscsi_login_rsp_hdr *rsp = (struct iscsi_login_rsp_hdr *)&conn->rsp.bhs;
+	struct iscsi_login_req_hdr *req =
+				(struct iscsi_login_req_hdr *)&conn->req.bhs;
+
 	char *name, *alias, *session_type, *target_name;
 	struct sockaddr_storage ss;
 	socklen_t slen = sizeof(struct sockaddr_storage);
@@ -399,17 +432,13 @@ static void login_start(struct connection *conn)
 	conn->cid = be16_to_cpu(req->cid);
 	conn->sid.id64 = req->sid.id64;
 	if (!conn->sid.id64) {
-		rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-		rsp->status_detail = ISCSI_STATUS_MISSING_FIELDS;
-		conn->state = STATE_EXIT;
+		login_rsp_ini_err(conn, ISCSI_STATUS_MISSING_FIELDS);
 		return;
 	}
 
 	name = text_key_find(conn, "InitiatorName");
 	if (!name) {
-		rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-		rsp->status_detail = ISCSI_STATUS_MISSING_FIELDS;
-		conn->state = STATE_EXIT;
+		login_rsp_ini_err(conn, ISCSI_STATUS_MISSING_FIELDS);
 		return;
 	}
 	conn->initiator = strdup(name);
@@ -420,52 +449,59 @@ static void login_start(struct connection *conn)
 	conn->auth_method = -1;
 	conn->session_type = SESSION_NORMAL;
 
-	getsockname(conn->fd, (struct sockaddr *) &ss, &slen);
+	if (getsockname(conn->fd, (struct sockaddr *) &ss, &slen))
+		log_error("%s: Failed to get local address.", __FUNCTION__);
 
 	if (session_type) {
 		if (!strcmp(session_type, "Discovery"))
 			conn->session_type = SESSION_DISCOVERY;
 		else if (strcmp(session_type, "Normal")) {
-			rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-			rsp->status_detail = ISCSI_STATUS_INV_SESSION_TYPE;
-			conn->state = STATE_EXIT;
+			login_rsp_ini_err(conn, ISCSI_STATUS_INV_SESSION_TYPE);
 			return;
 		}
 	}
 
 	if (conn->session_type == SESSION_NORMAL) {
 		if (!target_name) {
-			rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-			rsp->status_detail = ISCSI_STATUS_MISSING_FIELDS;
-			conn->state = STATE_EXIT;
+			login_rsp_ini_err(conn, ISCSI_STATUS_MISSING_FIELDS);
 			return;
 		}
 
-		if (!(conn->tid = target_find_by_name(target_name))
-			|| !cops->initiator_allow(conn->tid, conn->fd, name)
-			|| !cops->target_allow(conn->tid, (struct sockaddr *) &ss)
-			|| !isns_scn_allow(conn->tid, name)) {
-			rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-			rsp->status_detail = ISCSI_STATUS_TGT_NOT_FOUND;
-			conn->state = STATE_EXIT;
+		struct target * const target =
+			target_find_by_name(target_name);
+
+		if (!target
+		    || !cops->initiator_allow(target->tid, conn->fd, name)
+		    || !cops->target_allow(target->tid, (struct sockaddr *) &ss)
+		    || !isns_scn_allow(target->tid, name)) {
+			login_rsp_ini_err(conn, ISCSI_STATUS_TGT_NOT_FOUND);
 			return;
 		}
 
-/* 		if (conn->target->max_sessions && */
-/* 		    (++conn->target->session_cnt > conn->target->max_sessions)) { */
-/* 			conn->target->session_cnt--; */
-/* 			rsp->status_class = ISCSI_STATUS_INITIATOR_ERR; */
-/* 			rsp->status_detail = ISCSI_STATUS_TOO_MANY_CONN; */
-/* 			conn->state = STATE_EXIT; */
-/* 			return; */
-/* 		} */
+		if (target_redirected(target, conn, (struct sockaddr *) &ss)) {
+			log_debug(1, "redirected target %s login to %s:%s",
+					target->name, target->redirect.addr,
+					strlen(target->redirect.port) ?
+					target->redirect.port : "3260");
+
+			login_rsp_err(conn, ISCSI_STATUS_REDIRECT, target->redirect.type);
+			return;
+		}
+
+		conn->tid = target->tid;
+
+		if (target->max_nr_sessions &&
+		    (++target->nr_sessions > target->max_nr_sessions)) {
+			--target->nr_sessions;
+			log_debug(1, "rejecting session for target '%s': "
+				  "too many sessions", target_name);
+			login_rsp_ini_err(conn, ISCSI_STATUS_TOO_MANY_CONN);
+			return;
+		}
 
 		if (ki->param_get(conn->tid, 0, key_session,
-				  conn->session_param)) {
-			rsp->status_class = ISCSI_STATUS_TARGET_ERR;
-			rsp->status_detail = ISCSI_STATUS_SVC_UNAVAILABLE;
-			conn->state = STATE_EXIT;
-		}
+				  conn->session_param))
+			login_rsp_ini_err(conn, ISCSI_STATUS_SVC_UNAVAILABLE);
 	}
 	conn->exp_cmd_sn = be32_to_cpu(req->cmd_sn);
 	log_debug(1, "exp_cmd_sn: %d,%d", conn->exp_cmd_sn, req->cmd_sn);
@@ -476,8 +512,11 @@ static void login_finish(struct connection *conn)
 {
 	switch (conn->session_type) {
 	case SESSION_NORMAL:
-		if (!conn->session)
-			session_create(conn);
+		/* create or re-create in case the session closed */
+		if (session_create(conn)) {
+			login_rsp_tgt_err(conn, ISCSI_STATUS_TARGET_ERROR);
+			return;
+		}
 		conn->sid = conn->session->sid;
 		break;
 	case SESSION_DISCOVERY:
@@ -553,12 +592,11 @@ static void cmnd_exec_login(struct connection *conn)
 	rsp->max_version = ISCSI_VERSION;
 	rsp->active_version = ISCSI_VERSION;
 	rsp->itt = req->itt;
+	rsp->sid = req->sid;
 
 	if (/*req->max_version < ISCSI_VERSION ||*/
 	    req->min_version > ISCSI_VERSION) {
-		rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-		rsp->status_detail = ISCSI_STATUS_NO_VERSION;
-		conn->state = STATE_EXIT;
+		login_rsp_ini_err(conn, ISCSI_STATUS_NO_VERSION);
 		return;
 	}
 
@@ -698,22 +736,13 @@ static void cmnd_exec_login(struct connection *conn)
 	rsp->max_cmd_sn = cpu_to_be32(conn->max_cmd_sn);
 	return;
 init_err:
-	rsp->flags = 0;
-	rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-	rsp->status_detail = ISCSI_STATUS_INIT_ERR;
-	conn->state = STATE_EXIT;
+	login_rsp_ini_err(conn, ISCSI_STATUS_INIT_ERR);
 	return;
 auth_err:
-	rsp->flags = 0;
-	rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
-	rsp->status_detail = ISCSI_STATUS_AUTH_FAILED;
-	conn->state = STATE_EXIT;
+	login_rsp_ini_err(conn, ISCSI_STATUS_AUTH_FAILED);
 	return;
- target_err:
-	rsp->flags = 0;
-	rsp->status_class = ISCSI_STATUS_TARGET_ERR;
-	rsp->status_detail = ISCSI_STATUS_TARGET_ERROR;
-	conn->state = STATE_EXIT;
+target_err:
+	login_rsp_tgt_err(conn, ISCSI_STATUS_TARGET_ERROR);
 	return;
 }
 
@@ -770,8 +799,10 @@ static void cmnd_exec_text(struct connection *conn)
 	}
 
 	if (list_empty(&conn->rsp_buf_list) ||
-	    list_length_is_one(&conn->rsp_buf_list))
+	    list_length_is_one(&conn->rsp_buf_list)) {
 		rsp->flags = ISCSI_FLG_FINAL;
+		conn->ttt = ISCSI_RESERVED_TAG;
+	}
 
 	rsp->ttt = conn->ttt;
 
@@ -813,7 +844,7 @@ int cmnd_execute(struct connection *conn)
 		}
 		cmnd_exec_login(conn);
 		login_rsp = (struct iscsi_login_rsp_hdr *) &conn->rsp.bhs;
-		if (login_rsp->status_class)
+		if (login_rsp->status_class && login_rsp->status_class != ISCSI_STATUS_REDIRECT)
 			conn_free_rsp_buf_list(conn);
 		break;
 	case ISCSI_OP_TEXT_CMD:
