@@ -60,7 +60,7 @@ static char *rxbuf;
 static uint16_t transaction;
 static uint32_t current_timeout = 30; /* seconds */
 static char eid[ISCSI_NAME_LEN];
-static uint8_t ip[16]; /* IET supoprts only one portal */
+static uint8_t ip[16]; /* IET supports only one portal */
 static struct sockaddr_storage ss;
 
 int isns_scn_allow(uint32_t tid, char *name)
@@ -85,25 +85,30 @@ static int isns_get_ip(int fd)
 {
 	int err, i;
 	uint32_t addr;
-	struct sockaddr_storage lss;
-	socklen_t slen = sizeof(lss);
+	union {
+		struct sockaddr s;
+		struct sockaddr_storage ss;
+		struct sockaddr_in s4;
+		struct sockaddr_in6 s6;
+	} l;
+	socklen_t slen = sizeof(l.s);
 
-	err = getsockname(fd, (struct sockaddr *) &lss, &slen);
+	err = getsockname(fd, &l.s, &slen);
 	if (err) {
 		log_error("getsockname error %s!", gai_strerror(err));
 		return err;
 	}
 
-	err = getnameinfo((struct sockaddr *) &lss, sizeof(lss),
+	err = getnameinfo(&l.s, sizeof(l.s),
 			  eid, sizeof(eid), NULL, 0, 0);
 	if (err) {
 		log_error("getaddrinfo error %s!", gai_strerror(err));
 		return err;
 	}
 
-	switch (lss.ss_family) {
+	switch (l.ss.ss_family) {
 	case AF_INET:
-		addr = (((struct sockaddr_in *) &lss)->sin_addr.s_addr);
+		addr = ((&l.s4)->sin_addr.s_addr);
 
 		ip[10] = ip[11] = 0xff;
 		ip[15] = 0xff & (addr >> 24);
@@ -113,7 +118,7 @@ static int isns_get_ip(int fd)
 		break;
 	case AF_INET6:
 		for (i = 0; i < ARRAY_SIZE(ip); i++)
-			ip[i] = ((struct sockaddr_in6 *) &lss)->sin6_addr.s6_addr[i];
+			ip[i] = (&l.s6)->sin6_addr.s6_addr[i];
 		break;
 	}
 
@@ -442,19 +447,10 @@ static void free_all_acl(struct target *target)
 	while (!list_empty(&target->isns_head)) {
 		ini = list_entry(target->isns_head.q_forw, typeof(*ini), ilist);
 		remque(&ini->ilist);
+		free(ini);
 	}
 }
 
-static struct target *target_lookup_by_name(char *name)
-{
-	uint32_t tid;
-
-	tid = target_find_by_name(name);
-	if (!tid)
-		return NULL;
-	return target_find_by_id(tid);
-}
-
 int isns_target_deregister(char *name)
 {
 	char buf[4096];
@@ -464,7 +460,7 @@ int isns_target_deregister(char *name)
 	int err, last = list_empty(&targets_list);
 	struct target *target;
 
-	target = target_lookup_by_name(name);
+	target = target_find_by_name(name);
 	if (target)
 		free_all_acl(target);
 
@@ -605,18 +601,36 @@ static char *print_scn_pdu(struct isns_hdr *hdr)
 	while (length) {
 		uint32_t vlen = ntohl(tlv->length);
 
+		if (vlen + sizeof(*tlv) > length)
+			vlen = length - sizeof(*tlv);
+
 		switch (ntohl(tlv->tag)) {
 		case ISNS_ATTR_ISCSI_NAME:
-			log_error("scn name: %u, %s", vlen, (char *) tlv->value);
-			if (!name)
-				name = (char *) tlv->value;
+			if (vlen) {
+				size_t slen = vlen - 1;
+
+				if (slen > ISCSI_NAME_LEN)
+					slen = ISCSI_NAME_LEN;
+
+				*((char *) tlv->value + slen) = 0;
+
+				log_error("scn name: %u, %s", vlen,
+					  (char *) tlv->value);
+
+				if (!name)
+					name = (char *) tlv->value;
+			}
 			break;
 		case ISNS_ATTR_TIMESTAMP:
-/* 			log_error("%u : %u : %" PRIx64, ntohl(tlv->tag), vlen, */
-/* 				  *((uint64_t *) tlv->value)); */
+/*			if (vlen == 8)
+				log_error("%u : %u : %" PRIx64, ntohl(tlv->tag),
+					  vlen, *((uint64_t *) tlv->value));
+*/
 			break;
 		case ISNS_ATTR_ISCSI_SCN_BITMAP:
-			log_error("scn bitmap : %x", *((uint32_t *) tlv->value));
+			if (vlen == 4)
+				log_error("scn bitmap : %x",
+					  *((uint32_t *) tlv->value));
 			break;
 		}
 
@@ -665,7 +679,7 @@ found:
 		goto free_qry_mgmt;
 	}
 
-	target = target_lookup_by_name(mgmt->name);
+	target = target_find_by_name(mgmt->name);
 	if (!target) {
 		log_error("%s %d: invalid tid %s",
 			  __FUNCTION__, __LINE__, mgmt->name);
@@ -676,24 +690,45 @@ found:
 
 	/* skip status */
 	tlv = (struct isns_tlv *) ((char *) hdr->pdu + 4);
+
+	if (length < 4)
+		goto free_qry_mgmt;
+
 	length -= 4;
 
 	while (length) {
 		uint32_t vlen = ntohl(tlv->length);
 
+		if (vlen + sizeof(*tlv) > length)
+			vlen = length - sizeof(*tlv);
+
 		switch (ntohl(tlv->tag)) {
 		case ISNS_ATTR_ISCSI_NAME:
-			name = (char *) tlv->value;
+			if (vlen) {
+				size_t slen = vlen - 1;
+
+				if (slen > ISCSI_NAME_LEN)
+					slen = ISCSI_NAME_LEN;
+
+				*((char *) tlv->value + slen) = 0;
+
+				name = (char *) tlv->value;
+			} else
+				name = NULL;
 			break;
 		case ISNS_ATTR_ISCSI_NODE_TYPE:
-			if (ntohl(*(tlv->value)) == ISNS_NODE_INITIATOR && name) {
+			if (vlen == 4 && name &&
+			    ntohl(*(tlv->value)) == ISNS_NODE_INITIATOR) {
 				log_error("%s %d: %s", __FUNCTION__, __LINE__,
 					  (char *) name);
+
 				ini = malloc(sizeof(*ini));
 				if (!ini)
 					goto free_qry_mgmt;
+
 				snprintf(ini->name, sizeof(ini->name), "%s",
 					 name);
+
 				insque(&ini->ilist, &target->isns_head);
 			} else
 				name = NULL;
@@ -858,7 +893,12 @@ int isns_scn_handle(int is_accept)
 static int scn_init(char *addr)
 {
 	int fd, opt, err;
-	struct sockaddr_storage lss;
+	union {
+		struct sockaddr s;
+		struct sockaddr_storage ss;
+		struct sockaddr_in s4;
+		struct sockaddr_in6 s6;
+	} l;
 	socklen_t slen;
 
 	fd = socket(ss.ss_family, SOCK_STREAM, IPPROTO_TCP);
@@ -882,18 +922,18 @@ static int scn_init(char *addr)
 		goto out;
 	}
 
-	slen = sizeof(lss);
-	err = getsockname(fd, (struct sockaddr *) &lss, &slen);
+	slen = sizeof(l.s);
+	err = getsockname(fd, &l.s, &slen);
 	if (err) {
 		log_error("%s %d: %s\n", __FUNCTION__, __LINE__, strerror(errno));
 		goto out;
 	}
 
 	/* protocol independent way ? */
-	if (lss.ss_family == AF_INET6)
-		scn_listen_port = ntohs(((struct sockaddr_in6 *) &lss)->sin6_port);
+	if (l.ss.ss_family == AF_INET6)
+		scn_listen_port = ntohs((&l.s6)->sin6_port);
 	else
-		scn_listen_port = ntohs(((struct sockaddr_in *) &lss)->sin_port);
+		scn_listen_port = ntohs((&l.s4)->sin_port);
 
 	log_error("scn listen port %u %d %d\n", scn_listen_port, fd, err);
 out:
@@ -921,7 +961,7 @@ int isns_init(char *addr, int isns_ac)
 		log_error("getaddrinfo error %s, %s", gai_strerror(err), addr);
 		return -1;
 	}
-	memcpy(&ss, res->ai_addr, sizeof(ss));
+	memcpy(&ss, res->ai_addr, sizeof(*res->ai_addr));
 	freeaddrinfo(res);
 
 	rxbuf = calloc(2, BUFSIZE);
